from __future__ import annotations

from functools import singledispatch
from pathlib import Path
from typing import Any, Iterable, NoReturn, Sequence

import libcst as cst
import libcst.matchers as m

from cstq.csttraformers import ReplaceNodeTransformer
from cstq.cstvisitors import Extractor
from cstq.matchers import matcher
from cstq.node2id import NodeIDProvider


class CollectionOfNodes:
    def __init__(self, node_ids: Iterable[str], root: Query) -> None:
        self.__node_ids = sorted(set(node_ids))
        self.root = root

    @property
    def __nodes(self) -> dict[str, cst.CSTNode]:
        return {node_id: node for node_id, node in self.root.get_nodes_by_id(self.__node_ids).items() if node}

    @property
    def _nodes_id(self) -> list[str]:
        return self.__node_ids

    def filter(self, test) -> CollectionOfNodes:  # noqa: A003
        match_test = matcher(test)
        return CollectionOfNodes(
            [node_id for node_id in self.__node_ids if match_test(self.root.get_node_by_id(node_id))],
            root=self.root,
        )

    def filter_by_type(self, type_, test=None) -> CollectionOfNodes:
        return self.filter(lambda x: isinstance(x, type_) and (test(x) if test is not None else True))

    def search(self, test) -> CollectionOfNodes:
        return CollectionOfNodes(
            (
                self.root.get_node_id(found_node)
                for node in self.__nodes.values()
                for found_node in Extractor.match(node, matcher(test))
            ),
            root=self.root,
        )

    def search_by_type(self, type_, test=None) -> CollectionOfNodes:
        return self.search(lambda x: isinstance(x, type_) and (test(x) if test is not None else True))

    # find functions
    def find_function_call(self, func_name: str | None = None) -> CollectionOfNodes:
        nodes = self.search(m.Call(func=m.Name(value=func_name) if func_name else m.DoNotCare()))
        return nodes

    def __or__(self, other):
        assert self.root == other.root, "Both collections must have the same root"
        CollectionOfNodes([*self.__node_ids, *other._nodes_id], root=self.root)

    def __getitem__(self, item) -> CollectionOfNodes:
        if callable(item):
            return self.filter(item)
        if isinstance(item, int):
            return CollectionOfNodes([self.__node_ids[item]], root=self.root)
        if isinstance(item, slice):
            # maybe treat slice a bit different
            return CollectionOfNodes(self.__node_ids[item], root=self.root)

        msg = f"{item} of {type(item)=} not implemented/allowed"
        raise NotImplementedError(msg)

    def __getattr__(self, item) -> CollectionOfNodes:
        results: list[cst.CSTNode] = []
        for _node_id, node in self.__nodes.items():
            if not hasattr(node, item):
                continue
            attr = getattr(node, item)
            if isinstance(attr, (list, tuple)):
                results.extend(attr)
            else:
                results.append(attr)

        return CollectionOfNodes(self.root.get_nodes_id(results).values(), self.root)

    def __repr__(self) -> str:
        return f"<CollectionOfNodes nodes={self.__node_ids}>"

    def __len__(self) -> int:
        return len(self.__node_ids)

    def change(self, **kwargs) -> CollectionOfNodes:
        return CollectionOfNodes(
            self.root.replace_nodes({node_id: node.with_changes(**kwargs) for node_id, node in self.__nodes.items()}),
            self.root,
        )

    def replace(self, new_node: cst.CSTNode | cst.RemovalSentinel) -> CollectionOfNodes:
        new_node_ids = self.root.replace_nodes({node_id: new_node for node_id in self.__node_ids})
        return CollectionOfNodes([new_id for new_id in new_node_ids if new_id], self.root)

    # def swap(self, new_node: cst.CSTNode )  -> CollectionOfNodes:
    #     assert len(self.__node_ids) == 1, "you can only swap one node"
    #
    #     self.root.replace_nodes({
    #         self.__node_ids[0]: new_node,
    #         self.root.get_node_id(new_node): self.node(),
    #     })

    def remove(self) -> CollectionOfNodes:
        return self.replace(cst.RemovalSentinel.REMOVE)

    def node(self) -> cst.CSTNode:
        assert len(self.__node_ids) <= 1, "you have more than one, use nodes to get them all"
        assert len(self.__node_ids) == 1, "you have 0 nodes"

        return self.__nodes[self.__node_ids[0]]

    def nodes(self) -> list[cst.CSTNode]:
        return [*self.__nodes.values()]

    def childrens(self) -> CollectionOfNodes:
        node_ids = self.root.get_nodes_id(
            [children for node in self.__nodes.values() for children in node.children]
        ).values()
        return CollectionOfNodes(list(node_ids), self.root)

    def parent(self) -> CollectionOfNodes:
        node_ids = self.root.get_nodes_id(
            [self.root.get_parent_of_node(node) for node in self.__nodes.values()]
        ).values()
        return CollectionOfNodes(list(node_ids), self.root)

    def search_for_parents(self, test) -> CollectionOfNodes:
        result = []
        parents = {self.root.get_parent_of_node(node) for node in self.__nodes.values()}
        while len(parents) != 1 and not isinstance([*parents][0], cst.Module):
            for parent in parents:
                if test(parent):
                    result.append(self.root.get_node_id(parent))
            parents = {self.root.get_parent_of_node(node) for node in parents}
        return CollectionOfNodes(list(set(result)), self.root)


class Query(CollectionOfNodes):
    def __init__(self, mod: cst.CSTNode | str | Path) -> None:
        parsed_mod: cst.Module = parse_module(mod)
        self.wrapper: cst.metadata.MetadataWrapper = cst.metadata.MetadataWrapper(parsed_mod)
        self.module: cst.Module = self.wrapper.module

        self.__node_to_id: dict[cst.CSTNode, str] = self.wrapper.resolve(NodeIDProvider)
        self.__id_to_node: dict[str, cst.CSTNode] = {v: k for k, v in self.__node_to_id.items()}

        CollectionOfNodes.__init__(self, [self.get_node_id(self.module)], self)

    def get_node_id(self, node: cst.CSTNode) -> str:
        return self.__node_to_id[node]

    def get_nodes_id(self, nodes: list[cst.CSTNode]) -> dict[cst.CSTNode, str]:
        return {node: self.get_node_id(node) for node in nodes}

    def get_node_by_id(self, node_id: str) -> cst.CSTNode:
        return self.__id_to_node[node_id]

    def get_nodes_by_id(self, ids: Sequence[str]) -> dict[str, cst.CSTNode]:
        return {id_: self.get_node_by_id(id_) for id_ in ids}

    def replace_nodes(self, replace_map: dict[str, cst.CSTNode | cst.RemovalSentinel]):
        mod = self.wrapper.visit(ReplaceNodeTransformer(replace_map))
        self.__init__(mod)  # type: ignore

        return [None if to_node == cst.RemovalSentinel.REMOVE else from_id for from_id, to_node in replace_map.items()]

    def get_parent_of_node(self, node: cst.CSTNode) -> cst.CSTNode:
        return self.wrapper.resolve(cst.metadata.ParentNodeProvider)[node]

    def code(self) -> str:
        return self.module.code

    def write(self, path: Path) -> int:
        return path.write_text(self.code())


@singledispatch
def parse_module(module: Any) -> NoReturn:
    msg = f"Cant parse {type(module)} as module"
    raise RuntimeError(msg)


@parse_module.register(cst.Module)
def _(module: cst.Module) -> cst.Module:
    return module


@parse_module.register(str)
def _(module: str) -> cst.Module:
    return parse_module(cst.parse_module(module))


@parse_module.register(Path)
def _(module: Path) -> cst.Module:
    return parse_module(module.read_text())



import libcst as cst
import pytest

from cstq import Query


def test_init_works_with_string():
    q = Query(
        mod_text := """
# this is a random module
one = 1
two = one + one
    """.strip()
    )
    assert mod_text == q.code()


def test_init_works_with_path(tmp_path):
    test_file = tmp_path / "module.py"
    test_file.write_text(
        mod_text := """
# this is a random module
one = 1
two = one + one
    """.strip()
    )
    assert mod_text == Query(test_file).code()


def test_init_works_with_cstmodule():
    test_module = cst.parse_module(
        mod_text := """
# this is a random module
one = 1
two = one + one
    """.strip()
    )
    assert mod_text == Query(test_module).code()


def test_init_fail_with_non_module_node():
    with pytest.raises(RuntimeError) as exc_info:
        Query(cst.SimpleString(value="'value'"))
    assert str(exc_info.value) == "Cant parse <class 'libcst._nodes.expression.SimpleString'> as module"


def test_write(tmp_path):
    test_file = tmp_path / "dest.py"
    Query(
        mod_text := """
# this is a random module
one = 1
two = one + one
    """.strip()
    ).write(test_file)
    assert test_file.read_text() == mod_text



from typing import Callable

import libcst as cst


class Extractor(cst.CSTVisitor):
    def __init__(self, match_test: Callable[[cst.CSTNode], bool]) -> None:
        self.match_test = match_test
        self.result: list[cst.CSTNode] = []
        super().__init__()

    def on_visit(self, node: cst.CSTNode) -> bool:
        if self.match_test(node):
            self.result.append(node)
        return True

    @staticmethod
    def match(node: cst.CSTNode, match_test: Callable[[cst.CSTNode], bool]) -> list[cst.CSTNode]:
        """
        returns all the children nodes (counting node itself) that pass the match test
        """
        e = Extractor(match_test)
        node.visit(e)
        return e.result



from functools import partial
from typing import Callable

import libcst as cst
import libcst.matchers as m


def match(
    node: cst.CSTNode,
    test: m.BaseMatcherNode | Callable[[cst.CSTNode], bool] | cst.CSTNode,
) -> bool:
    if isinstance(test, m.BaseMatcherNode):
        return m.matches(node, test)
    elif callable(test):
        return test(node)
    elif isinstance(test, cst.CSTNode):
        return node.deep_equals(test)
    else:
        return test == node


def matcher(test: m.BaseMatcherNode | Callable[[cst.CSTNode], bool] | cst.CSTNode) -> Callable[[cst.CSTNode], bool]:
    return partial(match, test=test)



import pytest


@pytest.fixture()
def simple_bzl():
    """
    returns a simple bzl file that can be used for testing
    """
    return """
load("@build//:rust_binary.bzl", "rust_binary")
load("@build//:rust_library.bzl", "rust_library", "rust_library_helpers")
load("@tools//build_defs:glob_defs.bzl", "glob")

oncall("there_is_no_oncall")

rust_binary(
  name='greet',
  srcs=[
    'greet.rs',
  ],
  deps=[
    ':greeting',
  ],
)

rust_library(
  name='greeting',
  srcs=[
    'greeting.rs',
  ],
  deps=[
    ':join',
  ],
)

rust_library(
  name='join',
  srcs=[
    'join.rs',
  ],
)
    """.strip()



import libcst as cst

from cstq.node2id import NodeIDProvider


class ReplaceNodeTransformer(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (NodeIDProvider,)

    def __init__(self, replace_map: dict[str, cst.CSTNode | cst.RemovalSentinel]) -> None:
        self.replace_map = replace_map
        super().__init__()

    def on_leave(  # type: ignore
        self, old_node: cst.CSTNodeT, new_node: cst.CSTNode | cst.RemovalSentinel
    ) -> cst.CSTNode | cst.RemovalSentinel:
        node_id = self.get_metadata(NodeIDProvider, old_node)
        node = self.replace_map.get(node_id)

        if node:
            return node

        return new_node



from functools import singledispatch
from typing import Generator, Tuple

import libcst as cst


class NodeIDProvider(cst.VisitorMetadataProvider[str]):
    def visit_Module(self, module_node: cst.Module) -> None:  # noqa: N802
        for node, node_id in node2id(module_node, "$"):
            self.set_metadata(node, node_id)


@singledispatch
def node2id(
    node,  # noqa: ARG001
    id_prefix: str,  # noqa: ARG001
) -> Generator[Tuple[cst.CSTNode, str], None, None]:
    yield from []


@node2id.register(list)
@node2id.register(set)
@node2id.register(tuple)
def _(e: list, id_prefix: str) -> Generator[Tuple[cst.CSTNode, str], None, None]:
    for n, elem in enumerate(e):
        yield from node2id(elem, f"{id_prefix}[{n}]")


@node2id.register(cst.CSTNode)
def _(node, id_prefix: str) -> Generator[Tuple[cst.CSTNode, str], None, None]:
    id_prefix = f"{id_prefix}({type(node).__name__})"
    yield node, id_prefix

    for attr in node.__dataclass_fields__:
        yield from [inner for inner in node2id(getattr(node, attr), id_prefix=f"{id_prefix}.{attr}") if inner]



import libcst as cst
import libcst.matchers as m

from cstq import Query

MODULE = """
from foo import bar, baz
from f00 import b4r

one = 1

bar()
baz()
b4r("something")
bar()
baz()
""".strip()


def test_repalce_funtion():
    q = Query(MODULE)
    q.find_function_call(func_name="baz").parent().remove()
    assert (
        q.code()
        == """
from foo import bar, baz
from f00 import b4r

one = 1

bar()
b4r("something")
bar()
""".strip()
    )


def test_repalce_funtion_argument():
    q = Query(MODULE)
    q.find_function_call(func_name="b4r").args[0].value.change(value='"other"')
    assert (
        q.code()
        == """
from foo import bar, baz
from f00 import b4r

one = 1

bar()
baz()
b4r("other")
bar()
baz()
""".strip()
    )


def test_repalce_funtion_argument_if_exists():
    q = Query(MODULE)
    fcall = q.find_function_call(func_name="bar")[0]
    if len(fcall.args) == 0:
        fcall.change(args=(cst.Arg(value=cst.Name(value="noargs")),))

    assert (
        q.code()
        == """
from foo import bar, baz
from f00 import b4r

one = 1

bar(noargs)
baz()
b4r("something")
bar()
baz()
""".strip()
    )


def test_remove_funct(simple_bzl):
    q = Query(simple_bzl)

    load_glob = q.search(
        m.Call(
            func=m.Name(value="load"),
            args=[
                m.Arg(value=m.SimpleString(value='"@tools//build_defs:glob_defs.bzl"')),
                m.ZeroOrMore(),
                m.Arg(value=m.SimpleString(value='"glob"')),
                m.ZeroOrMore(),
            ],
        )
    )
    args = load_glob.args
    assert len(args) > 1
    args = args[1:]
    if len(args) == 1:
        load_glob.parent().remove()
    else:
        args.filter(m.Arg(value=m.SimpleString(value='"glob"'))).remove()

    assert q.code() == simple_bzl.replace("""load("@tools//build_defs:glob_defs.bzl", "glob")\n""", "")


def test_change_oncall(simple_bzl):
    q = Query(simple_bzl)
    q.find_function_call(func_name="oncall").args[0].value.change(value='"new_oncall"')
    assert q.code() == simple_bzl.replace("there_is_no_oncall", "new_oncall")



from cstq import Query

MODULE = """
# this is a random module
one = 1
two = one + one
""".strip()


def test_get_node_by_id():
    """
    This test is here to prove that we can get the main module by id.
    and that getting a node by id, and then asking the id of the node are idemponent actions
    """
    q = Query(MODULE)
    module_id = "$(Module)"
    module = q.get_node_by_id(module_id)
    assert module == q.module
    assert q.get_node_id(q.get_node_by_id(module_id)) == module_id


def test_get_nodes_by_id():
    """
    This test is here to prove that we can get the main module by id.
    and that getting a node by id, and then asking the id of the node are idemponent actions
    """
    q = Query(MODULE)
    module = q.module

    simple_statement_line = q.get_nodes_by_id(
        ["$(Module).body[0](SimpleStatementLine)", "$(Module).body[1](SimpleStatementLine)"]
    )

    assert simple_statement_line == {
        "$(Module).body[0](SimpleStatementLine)": module.body[0],
        "$(Module).body[1](SimpleStatementLine)": module.body[1],
    }


def test_get_nodes_id():
    q = Query(MODULE)
    module = q.module

    simple_statement_line = q.get_nodes_id([module.body[0], module.body[1]])

    assert simple_statement_line == {
        module.body[0]: "$(Module).body[0](SimpleStatementLine)",
        module.body[1]: "$(Module).body[1](SimpleStatementLine)",
    }


def test_get_parent():
    "makes sure that get_parent returns the node parent"
    q = Query(MODULE)
    module = q.module
    assert q.get_parent_of_node(module.body[1]) == module
